DAT objectName          byte "Nordic131126a", 0
CON
{{
  ******* Public Notes *******

  The PASM routines "targetIoOn"
  and "targetIoRoutineOff" leave the I/O
  pin pulled low for over 1/4 second.
  This allows the visual indication that
  the pin was triggered
  by attaching an LED to the "targetIo_"
  pin.

  If this object is used with a slave
  unit, the lines of code checking for
  the trigger button are replaced with
  "not an operation" statements "nop".
  This does not speed up the loop.  I
  may add code to both shorten the
  check IRQ loop.
  
}}
{
  ******* Private Notes *******

  131121a Start changing some variable names in
  preparation to forum release.
  22a Change set pipe routine to leave CE high at end.
  131123a Add Configure method.
  24a Add a pause to SetConfigure method. Why is one
  needed?
  24b Remove debug statements and methods.
  26a Delete many of these notes.
  
  
  Things to add:
  Decide how to handle data pipes.
  Should each data pipe have its own rx counter?
  Should each data pipe have its own rx buffer?
  Check for payload.  Make IRQ polling optional.
  Auto retransmit.
  Auto acknowledgement.
  Changing addresses (tx and rx) after initial configuation.
  FDX type interface. Probably not.  This would make things very complicated.
  Timed tx based on ID (anti-collision code).
  Message hopping.
  Bootloader / wireless programming.
  Interface with older nRF24A modules.
  Variable payload size?
  Timed tx/rx times.
  Fewer pins.  Optional IRQ?
    
}
CON
  
  
  'nRF24L01 constants:
{  
  'Commands
  R_REGISTER = %0000_0000 '+ reg
  W_REGISTER = %0010_0000 '+ reg  $20
  R_RX_PAYLOAD = %0110_0001     ' $61
  W_TX_PAYLOAD = %1010_0000     ' $A0
  FLUSH_TX = %1110_0001         ' $E1
  FLUSH_RX = %1110_0010         ' $E2
  REUSE_TX_PL = %1110_0011      ' $E3
  R_RX_PL_WID = %0110_0000      ' $30
  W_ACK_PAYLOAD = %1010_1000 '+ pipe  ' $A8
  W_TX_PAYLOAD_NOACK = %1011_0000 ' $B0
  NOOP = %1111_1111             ' $FF
                                
  'Registers
  CONFIG = $00   '%0000_1000 / $08
  EN_AA = $01  '%0011_1111  / $3F
  EN_RXADDR = $02  '%0000_0011
  SETUP_AW = $03   '%0000_0011
  SETUP_RETR = $04  '%0000_0011
  RF_CH = $05   '%0000_0010
  RF_SETUP = $06  '%0000_1110
  _STATUS = $07     '%0000_1110 / $E 
  OBSERVE_TX = $08  '%0000_0000
  RPD = $09   '%0000_0000 
  RX_ADDR_P0 = $0A '$E7E7E7E7E7
  RX_ADDR_P1 = $0B '$C2C2C2C2C2
  RX_ADDR_P2 = $0C  '$C3
  RX_ADDR_P3 = $0D  '$C4 
  RX_ADDR_P4 = $0E  '$C5 
  RX_ADDR_P5 = $0F  '$C6 
  TX_ADDR = $10    '$E7E7E7E7E7
  RX_PW_P0 = $11   '%0000_0000
  RX_PW_P1 = $12   '%0000_0000 
  RX_PW_P2 = $13   '%0000_0000 
  RX_PW_P3 = $14  '%0000_0000 
  RX_PW_P4 = $15  '%0000_0000 
  RX_PW_P5 = $16  '%0000_0000 
  FIFO_STATUS = $17 '%0001_0001 
  DYNPD = $1C   '%0000_0000
  FEATURE = $1D  '%0000_0000   }
  
CON

  WRITE_REGISTER = $20
  
  '' registers 
  #0, CONFIG, EN_AA, EN_RX_ADDR, SETUP_AW, SETUP_RETR, RF_CH, RF_SETUP, STATUS, {
    } OBSERVE_TX, RPD, RX_ADDR_P0 {5 bytes}, RX_ADDR_P1 {5 bytes}, RX_ADDR_P2, RX_ADDR_P3, {
    } RX_ADDR_P4, RX_ADDR_P5, TX_ADDR {5 bytes}, RX_PW_P0, RX_PW_P1, RX_PW_P2, RX_PW_P3, {
    } RX_PW_P4, RX_PW_P5, FIFO_STATUS
  DYNPD = $1C ' Enable dynamic payload length   
  FEATURE = $1D
  
  '' hub offset of register data 
  #0, CONFIG_OFFSET, EN_AA_OFFSET, EN_RX_ADDR_OFFSET, SETUP_AW_OFFSET, SETUP_RETR_OFFSET, {
    } RF_CH_OFFSET, RF_SETUP_OFFSET, STATUS_OFFSET, OBSERVE_TX_OFFSET, RPD_OFFSET, {
    } RX_ADDR_P0_OFFSET

  RX_ADDR_P1_OFFSET = RX_ADDR_P0_OFFSET + 5
  RX_ADDR_P2_OFFSET = RX_ADDR_P1_OFFSET + 5
  RX_ADDR_P3_OFFSET = RX_ADDR_P2_OFFSET + 1
  RX_ADDR_P4_OFFSET = RX_ADDR_P3_OFFSET + 1
  RX_ADDR_P5_OFFSET = RX_ADDR_P4_OFFSET + 1
  TX_ADDR_OFFSET = RX_ADDR_P5_OFFSET + 1
  RX_PW_P0_OFFSET = TX_ADDR_OFFSET + 5
  RX_PW_P1_OFFSET = RX_PW_P0_OFFSET + 1
  RX_PW_P2_OFFSET = RX_PW_P1_OFFSET + 1
  RX_PW_P3_OFFSET = RX_PW_P2_OFFSET + 1
  RX_PW_P4_OFFSET = RX_PW_P3_OFFSET + 1
  RX_PW_P5_OFFSET = RX_PW_P4_OFFSET + 1
  FIFO_STATUS_OFFSET = RX_PW_P5_OFFSET + 1
  
  DYNPD_OFFSET = FIFO_STATUS_OFFSET + 1 
  FEATURE_OFFSET = DYNPD_OFFSET + 1 

  R_RX_PAYLOAD = %0110_0001 
  W_TX_PAYLOAD = %1010_0000    
  FLUSH_TX = %1110_0001       
  FLUSH_RX = %1110_0010       
  REUSE_TX_PL = %1110_0011  
  R_RX_PL_WID = %0110_0000   
  W_ACK_PAYLOAD = %1010_1000 '+ pipe  *** learn to use this
  W_TX_PALOAD_NO = %1011_0000 ' disable AUTOACK on specific packet *** learn to use this
  NOT_AN_OPPERATION = $FF            

  #0, IDLE, CMD_READ, CMD_WRITE, READ_PAYLOAD, TRANSMIT_STRING, SET_PIPE, TRANSMIT_ON_PIPE, {
    } CONFIGURE, ENABLE_HIGH, ENABLE_LOW, _AddDelay

  LARGEST_PAYLOAD_SIZE = 32  
  POSSIBLE_DATAPIPES = 6
  LARGEST_POSSIBLE_ADDRESS = 5
  RX_FIFO_EMPTY = 7
  TRIGGER_DATAPIPE = 5
  
  DEBUG_BAUD = 115200
  _SmallestDelay = 20           ' it may be possible to reduce this to 10, I'm not sure
                                ' what the smallest waitcnt value can be in PASM.

  _ZeroDelayOverhead = 272      ' 68 instructions from IRQ low to target IO pin low
  _AddedDelayOverhead = 12
  
VAR

  long cog, commandHub
  'long valueOut
  'long valueIn, globalFlag
  long rxPayloadPtr, txPayloadPtr, bufferSize
  long rxCountsHub, rxCountHub[POSSIBLE_DATAPIPES], triggerCountHub, targetCountHub
  long parentCommand, dataPipeHub, statusHub
  long txAddressLocation[POSSIBLE_DATAPIPES]
  
  byte txAddress[POSSIBLE_DATAPIPES * LARGEST_POSSIBLE_ADDRESS]
  
DAT

txAddressDat     byte $E7[5]
{txAddress1}    byte $C2[5]
{txAddress2}    byte $C3, $C2[4] '$C2[4], $C3  ' maybe this should be byte  
{txAddress3}    byte $C4, $C2[4]
{txAddress4}    byte $C5, $C2[4]
{txAddress5}    byte $C6, $C2[4]
  
'OBJ

  'Debug : "FullDuplexSerialPlusMultiObj"

PUB Start(irq_, miso_, mosi_, clock_, csn_, ce_, triggerPin, targetIo_, configPtr_) : okay
'' Start SPI Engine - starts a cog
'' returns false if no cog available
'' Prior to calling "Start" the method "Init" should be called
'' to set pointer for tx and rx buffers.

  bytemove(@txAddress, @txAddressDat, POSSIBLE_DATAPIPES * LARGEST_POSSIBLE_ADDRESS)

  dataStatus := byte[configPtr_ + STATUS_OFFSET] 
  
  configPtr := configPtr_
  
  stop
  
  payloadSize := byte[configPtr + RX_PW_P0_OFFSET] ' **payloadSize should be changable based on which data pipe is used.
  bufferSize := payloadSize 
  statusReset := byte[configPtr + STATUS_OFFSET]
  ceToggleDelay := clkfreq / 100_000 ' 800        
  ceToggleDelay #>= 20
  rxPayloadPtrLoc := @rxPayloadPtr
  txPayloadPtrLoc := @txPayloadPtr
  bufferSizePtr := @bufferSize
  rxCountsPtr := @rxCountsHub
  rxCountPtr := @rxCountHub     ' rxCountHub arrau would store the number of rx packets received 
                                ' on individual data pipes.  rxCountHub is currently not used.
  triggerCountPtr := @triggerCountHub
  targetIoCountPtr := @targetCountHub
  parentCommandPtr := @parentCommand
  dataPipePtr := @dataPipeHub
  dataPipeHub := TRIGGER_DATAPIPE
  statusPtr := @statusHub
  txAddressLocationPtr := @txAddressLocation

  addressSize := byte[configPtr + SETUP_AW_OFFSET] + 2

  repeat result from 0 to POSSIBLE_DATAPIPES - 1
    txAddressLocation[result] := @txAddress + (result * LARGEST_POSSIBLE_ADDRESS)
  
  txAddressLocationCog := txAddressLocation[0]       ' *** temp 

  clockMask := 1 << clock_
  misoMask := 1 << miso_
  mosiMask := 1 << mosi_
  irqMask := 1 << irq_
  if triggerPin <> -1
    triggerMask := 1 << triggerPin
  else
    triggerMask := 0
  if targetIo_ <> -1
    triggerTargetMask := 1 << targetIo_
  else
    triggerTargetMask := 0
    
  csnMask := 1 << csn_
  ceMask := 1 << ce_

  commandPtr := @commandHub
  
  quarterSecond := clkfreq / 4

  okay := cog := cognew(@enter, @commandHub) + 1   

PUB Init(rxPayload_, txPayload_)
'' Call Init prior to calling Start so this object knows where to
'' store incoming data and where to find out going data.
'' This method allows the automatic writes and reads caused
'' by trigger presses or incomming data.

  rxPayloadPtr := rxPayload_
  txPayloadPtr := txPayload_

PUB stop
'' Stop SPI Engine - frees a cog

  if cog
     cogstop(cog~ - 1)
  commandHub~

PUB CmdRead(localCommand, localPtr, bytes)
'' The value of "localcommand" will be first written to the
'' Nordic module.  The bytes read will be written to hub
'' RAM at "localPtr".
'' Set localCommand to -1 to read only.

  parentCommand := localCommand
  rxPayloadPtr := localPtr
  bufferSize := bytes
  SetCommand(CMD_READ)
   
PUB CmdWrite(localCommand, localPtr, bytes)
'' The value of "localcommand" will be first written to the
'' Nordic module.  The bytes located in hub
'' RAM at "localPtr" will then be written to the Nordic
'' module.
'' Set localCommand to -1 to write from pointer location
'' without writing a command.

  parentCommand := localCommand
  txPayloadPtr := localPtr  ' ** reset pointer back to previous value?  
  bufferSize := bytes
  SetCommand(CMD_WRITE)
   
PUB TransmitString(localPtr)
'' Transmits packet at "localPtr" on the last data pipe set.

  txPayloadPtr := localPtr
  SetCommand(TRANSMIT_STRING)
                      
PUB SetDataPipe(localPipe, localPtr)
'' Set the data pipe for the Nordic module to use
'' without transmitting a packet.
'' If localPtr equals zero or -1, the
'' previous txPayloadPtr value be used.

  dataPipeHub := localPipe
  if localPtr <> 0 and localPtr <> -1
    txPayloadPtr := localPtr 
  SetCommand(SET_PIPE)
  
PUB TransmitStringOnDataPipe(localPipe, localPtr) | previousPipe
'' Transmits packet at "localPtr" on data pipe "localPipe".

  'previousPipe := dataPipeHub
  dataPipeHub := localPipe
  txPayloadPtr := localPtr
 
  SetCommand(TRANSMIT_ON_PIPE)
 
PUB EnableHigh

  SetCommand(ENABLE_HIGH)

PUB EnableLow

  SetCommand(ENABLE_LOW)

'PUB PowerUp
' The module can be powered up by useing the CmdWrite method and setting the config data
' with a high power up bit.

  'SetCommand(_PowerUp)

'PUB PowerDown

  'SetCommand(_PowerDown)
    
PUB ReadTriggerCounts

  result := triggerCountHub
    
PUB ReadTargetActivity

  result := targetCountHub
    
PUB ReadRxCounts

  result := rxCountsHub
    
{PUB ReadRxCount(localPipe)
'' Doesn't work yet.
  result := rxCountHub[localPipe]
}    
PUB ReadStatus
'' returns the last status byte read from
'' Nordic module.

  result := statusHub

PUB SetConfigure

   
  waitcnt(clkfreq / 80 + cnt) 
  'waitcnt(clkfreq / 40 + cnt) ' works
  SetCommand(CONFIGURE)

PRI SetCommand(cmd)

  commandHub := cmd              ''write command
  repeat while commandHub        ''wait for command to be cleared, signifying receipt

PUB GetObjectName

  result := @objectName
      
DAT           org

DAT enter               or      outa, mosiMask  ' Set pins used by driver
                        or      dira, mosiMask                
                        andn    outa, clockMask        
                        or      dira, clockMask                
                        andn    outa, misoMask             
                        andn    dira, misoMask
                        or      outa, triggerTargetMask ' active low on slave device             
                        or      dira, triggerTargetMask
                        andn    outa, ceMask    ' tx mode when high rx mode when low             
                        or      dira, ceMask
                        or      outa, csnMask   ' active low          
                        or      dira, csnMask
                     
                        tjz     triggerMask, #fixTriggerCheck  ' **** clean up the way the code gets modified
                        jmp     #configure_  '#loop
                        
DAT preloop             wrlong  zero, commandPtr              ' Make the calling method wait until done for now.  
loop                    test    triggerMask, ina wc     'trigger pin to 'C' flag ' for active low button (I'm using buttons on PPDB)
loopP1        if_nc     jmp     #trigger_   ' problem was here
                        rdlong  cogCmd, commandPtr              ''wait for command
                        test    irqMask, ina wc     'IRQ pin to 'C' flag ' for active low pin
              if_nc     jmp     #readPayload_ 
                        tjz     cogCmd, #loop
                        mov     jumpMode, cogCmd        ' I could save a long by reusing cogCmd as jumpMode             
                        add     jumpMode, #JumpTable
                        sub     jumpMode, #1              
                        jmp     jumpMode   ' indirect jump
                        
DAT JumpTable           jmp     #cmdRead_
                        jmp     #cmdWrite_
                        jmp     #readPayload_
                        jmp     #txString_
                        jmp     #setPipe_
                        jmp     #sendPacket_
                        jmp     #configure_
                        jmp     #enableHigh_
                        jmp     #enableLow_
                        jmp     #addDelay_
                        'call   #txStringRoutine  ' from trigger
                        'add    triggerCountCog, #1
                        'wrlong triggerCountCog, triggerCountPtr
                        'jmp    #preloop
{  #0, IDLE, CMD_READ, CMD_WRITE, READ_PAYLOAD, TRANSMIT_STRING, SET_PIPE, TRANSMIT_ON_PIPE, {
    } CONFIGURE, ENABLE_HIGH, ENABLE_LOW, _AddDelay
}                        
'################################################################################################################

'------------------------------------------------------------------------------------------------------------------------------

DAT cmdRead_            rdlong  cogCmd, parentCommandPtr
                        andn    outa, csnMask
                        call    #commandCheck
                        rdlong  address, rxPayloadPtrLoc
                        rdlong  byteCountdown, bufferSizePtr
                        call    #readBytesRoutine
                        or      outa, csnMask
                        
                        jmp     #preloop
                        
'------------------------------------------------------------------------------------------------------------------------------
' Shift out the command byte.  If command byte (read as a long) equals -1 then no command is written.

commandCheck            cmp     cogCmd, negativeOne wz
              if_z      jmp     #commandCheck_ret
                        mov     oByte, cogCmd   
                        call    #shiftOut
commandCheck_ret        ret
'------------------------------------------------------------------------------------------------------------------------------

DAT cmdWrite_           rdlong  cogCmd, parentCommandPtr
                        andn    outa, csnMask
                        call    #commandCheck
                        rdlong  address, txPayloadPtrLoc
                        rdlong  byteCountdown, bufferSizePtr
                        call    #writeBytesRoutine
                        or      outa, csnMask
                        
                        jmp     #preloop
'------------------------------------------------------------------------------------------------------------------------------
' **** Add memory of previous data pipe so the various parameters don't need to be reset.
' I think this can read from any data pipe.
DAT readPayload_        andn    outa, ceMask            ' stop receiving       
                        'rdlong  byteCountdown, bufferSizePtr
                        'mov     byteCountdown, payloadSize ' *** what if we want to change payload size?                      
                        mov     oByte, #R_RX_PAYLOAD      'R_RX_PAYLOAD = %0110_0001     ' $61 
                        
                        andn    outa, csnMask
                        call    #shiftIO                ' read status to write to hub later
                                                        ' iByte now holds status register's contentsc
                        and     iByte, #%1111_1111      ' ** why is this needed? check on it                          
                        cmp     iByte, targetDataPipeCog wz   
jumpTargetIo     if_z      jmp     #triggerDelayCode                ' The destination of this line may be changed by addDelay_ code.
returnPoint             mov     statusCog, iByte

                        mov     rxPipe, statusCog   
                        and     rxPipe, #%0000_1110 
                        shr     rxPipe, #1 
                        mov     address, configPtr         ' address to hold payload size hub address
                        add     address, #RX_PW_P0_OFFSET
                        add     address, rxPipe
                        rdbyte  rxPayload, address     ' read payload size of this data pipe from hub
                        mov     byteCountdown, rxPayload
                        rdlong  address, rxPayloadPtrLoc  ' find out where to write payload in hub
                          
                        call    #readBytesRoutine   ' read the rest of payload and write to hub
endOfPayload            or      outa, csnMask
                         
                        mov     oByte, #FLUSH_RX   ' clear rx buffer of Nordic chip
                        
                        andn    outa, csnMask         
                        call    #shiftOut
                        or      outa, csnMask
                        
                        mov     address, configPtr         ' ** is this needed?
                        add     address, #STATUS_OFFSET       ' register x byte address in hub
                        mov     activeRegister, #STATUS
                        call    #writeRegister
                        
                        or      outa, ceMask    'Start receiving
                        wrlong  statusCog, statusPtr    ' write status to hub 
                        add     rxCountsCog, #1          ' ** use separate counters for each data pipe
                        wrlong  rxCountsCog, rxCountsPtr

                        call    #targetIoRoutineOff
                     
                        jmp     #preloop        'Go wait for next command
'------------------------------------------------------------------------------------------------------------------------------
' The this section of code will use the current data pipe.
DAT txString_           call    #txStringRoutine
                        
                        jmp     #preloop                  
'------------------------------------------------------------------------------------------------------------------------------
DAT setPipe_            call    #setPipeRoutine
                        jmp     #preloop 
'------------------------------------------------------------------------------------------------------------------------------
DAT setPipeRoutine      rdLong dataPipeCog, dataPipePtr ' ** add check to see if this dataPipe is different than previous
                        mov     oByte, #WRITE_REGISTER
                        andn    outa, ceMask                 
                        andn    outa, csnMask           ' actively communicating with module
                        call    #shiftOut      
                        mov     oByte, configData
                        add     oByte, #1              ' PRX powered down ** add check so module only powers down when changing addresses
                        call    #shiftOut
                        or      outa, csnMask          ' ready to write registers

                        mov     address, configPtr 
                        add     address, #RX_PW_P0_OFFSET       
                        add     address, dataPipeCog
                        rdbyte  payloadSize, address  ' read payload for this pipe line
                        
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #TX_ADDR
                        andn    outa, csnMask           ' actively communicating
                        call    #shiftOut               ' sent write TX_ADDR command                        
                                                
                        mov     byteCountdown, dataPipeCog  ' prepare countdown to find tx address data in hub 5 byte per data pipe
                        mov     address, txAddressLocationCog

testForEnd              tjz     byteCountdown, #readTxAddrPtr ' if data pipe zero don't adjust the pointer

                        
addFive                 add     address, #5             ' I'm using the direct pointer to address and add 5                        
                        djnz    byteCountdown, #addFive

readTxAddrPtr           mov     byteCountdown, addressSize      ' address should now point to first byte of tx address
                                                                ' prepare countdown for number of bytes in address
                        call    #writeBytesRoutine   ' write address
                        or      outa, csnMask          ' ready to write registers                        
                        or      outa, ceMask                                

setPipeRoutine_ret      ret                         
'------------------------------------------------------------------------------------------------------------------------------
DAT sendPacket_         call    #setPipeRoutine
                        call    #txStringRoutine                        
                        jmp     #preloop                  
'------------------------------------------------------------------------------------------------------------------------------
DAT txStringRoutine     mov     address, configPtr         
                        add     address, #STATUS_OFFSET       ' register x byte address in hub
                        mov     activeRegister, #STATUS

                        andn    outa, ceMask
                        call    #writeRegister

                        andn    outa, csnMask           ' actively communicating
                        mov     oByte, #WRITE_REGISTER ' config register
                        call    #shiftOut
                        mov     oByte, configData
                        add     oByte, #%10            ' power up PTX mode
                        call    #shiftOut
                        or      outa, csnMask           ' stop communicating
                        
                        mov     oByte, #FLUSH_TX       ' **Think about times when I'd want to retransmit.
                        andn    outa, csnMask   
                        call    #shiftOut
                        or      outa, csnMask
                        
                        mov     oByte, #W_TX_PAYLOAD     
                        rdlong  address, txPayloadPtrLoc
                        andn    outa, csnMask   
                        call    #shiftOut
                        mov     byteCountdown, payloadSize ' payloadSize is determined by the data pipe used                     
                        call    #writeBytesRoutine
                        or      outa, csnMask
                        
                        or      outa, ceMask            ' set ce high
                        mov     tempDelay, cnt
                        add     tempDelay, ceToggleDelay
                        waitcnt tempDelay, zero
                      
                                
txStringRoutine_ret     ret
'------------------------------------------------------------------------------------------------------------------------------
DAT configure_          mov     address, configPtr        
                        rdbyte  configData, address

                        and     configData, #%0111_1100 ' strip power and tx/rx mode off byte

                        mov     oByte, #WRITE_REGISTER ' config register is zero so nothing needs to be added 
                                                       ' to write to config reg
                        andn    outa, ceMask                 
                        andn    outa, csnMask           ' actively communicating with module
                        call    #shiftOut      '5j* 5m
                        mov     oByte, configData
                        add     oByte, #1              ' PRX
                        call    #shiftOut
                        or      outa, csnMask           ' stop communicating
                        nop
                        nop
                        mov     byteCountdown, #6       ' write to registers 1 - 6
                        mov     activeRegister, #EN_AA
                        mov     address, configPtr
                        add     address, #EN_AA_OFFSET
                        
writeRegisters          call    #writeRegister
                        add     activeRegister, #1      ' ** is it okay to write registers from high to low?
                        add     address, #1      
                        djnz    byteCountdown, #writeRegisters
                        
                        mov     address, configPtr
                        add     address, #RX_PW_P0_OFFSET ' register x byte address in hub
                        rdbyte  payloadSize, address    ' payload size may be changed by
                                                        ' using TransmitStringOnDataPipe method
                        
                        'mov     pipe0Payload, payloadSize
                        mov     activeRegister, #RX_PW_P0  
                        mov     byteCountdown, #6       ' write to registers 1 - 6

dataPipePayloads        call    #writeRegister
                        add     activeRegister, #1      ' ** is it okay to write registers from high to low?
                        add     address, #1      
                        djnz    byteCountdown, #dataPipePayloads
                        
                        {' addressSize is now set before PASM cog is launched
readAddressSize         mov     address, configPtr
                        add     address, #SETUP_AW_OFFSET                             
                        rdbyte  addressSize, address    
                        add     addressSize, #2 }
                        mov     byteCountdown, addressSize
                        mov     address, configPtr
                        add     address, #RX_ADDR_P0_OFFSET
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #RX_ADDR_P0       
                        andn    outa, csnMask           ' actively communicating
                        call    #shiftOut
                        
                        call    #writeBytesRoutine
                        or      outa, csnMask           ' stop communicating

                       { nop   ' Leave RX address at their default values for now.
                        nop
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #RX_ADDR_P1
                        andn    outa, csnMask
                        call    #shiftOut
                        mov     byteCountdown, addressSize 
                        mov     address, txAddress1Ptr
                        call    #writeBytesRoutine
                        or      outa, csnMask
                        nop
                        nop
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #RX_ADDR_P2
                        andn    outa, csnMask
                        call    #shiftOut
                        mov     oByte, #$C3
                        call    #shiftOut
                        or      outa, csnMask
                        nop
                        nop
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #RX_ADDR_P3
                        andn    outa, csnMask
                        call    #shiftOut
                        mov     oByte, #$C4
                        call    #shiftOut
                        or      outa, csnMask
                        nop
                        nop
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #RX_ADDR_P4
                        andn    outa, csnMask
                        call    #shiftOut
                        mov     oByte, #$C5
                        call    #shiftOut
                        or      outa, csnMask
                        nop
                        nop
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #RX_ADDR_P5
                        andn    outa, csnMask
                        call    #shiftOut
                        mov     oByte, #$C6
                        call    #shiftOut
                        or      outa, csnMask        }
                        nop
                        nop

'readTxAddress
                        mov     address, configPtr
                        add     address, #TX_ADDR_OFFSET
                        mov     byteCountdown, addressSize  
                        mov     oByte, #WRITE_REGISTER
                        add     oByte, #TX_ADDR
                        andn    outa, csnMask           ' actively communicating
                        call    #shiftOut
                        
                        call    #writeBytesRoutine
                        or      outa, csnMask           ' stop communicating
                        nop
                        nop
                        mov     oByte, #WRITE_REGISTER
                        andn    outa, csnMask           ' actively communicating
                        call    #shiftOut
                        mov     oByte, configData
                        add     oByte, #%11 ' power up PRX mode
                        call    #shiftOut
                        or      outa, csnMask           ' stop communicating
                        nop
                        
                        'Start receiving
                        or      outa, ceMask                 
                        jmp     #preloop                'Go wait for command
'------------------------------------------------------------------------------------------------------------------------------
'' This section allows access to CE pin from Spin.
DAT enableHigh_         or      outa, ceMask                 
                        jmp     #preloop

'------------------------------------------------------------------------------------------------------------------------------
'' This section allows access to CE pin from Spin.
DAT enableLow_          or      outa, ceMask                 
                        jmp     #preloop
'------------------------------------------------------------------------------------------------------------------------------
DAT trigger_            call    #txStringRoutine        ' Send 
                        waitpeq triggerMask, triggerMask ' wait for trigger pin to go high again.
                        add     triggerCountCog, #1
                        wrlong  triggerCountCog, triggerCountPtr
                        jmp     #preloop
'------------------------------------------------------------------------------------------------------------------------------
DAT addDelay_           rdlong  triggerDelay, parentCommandPtr
                        'tjz     triggerDelay, #zeroDelayCoding
                        'movs    jumpTargetIo, #triggerDelayCode
                        jmp     #preloop
zeroDelayCoding         movs    jumpTargetIo, #targetIoOn
                        jmp     #preloop
'------------------------------------------------------------------------------------------------------------------------------
'' values of byteCountdown and address should be set before calling this subroutine
'' csn pin not controlled here
DAT readBytesRoutine    call    #shiftIn
                        wrbyte  iByte, address
                        add     address, #1
                        djnz    byteCountdown, #readBytesRoutine
                        
readBytesRoutine_ret    ret
'------------------------------------------------------------------------------------------------------------------------------
'' values of byteCountdown and address should be set before calling this subroutine

DAT writeBytesRoutine   rdbyte  oByte, address    
                        call    #shiftOut
                        add     address, #1             ' address byte address in hub
                        djnz    byteCountdown, #writeBytesRoutine

writeBytesRoutine_ret   ret
'------------------------------------------------------------------------------------------------------------------------------
'' values of activeRegister and address should be set before calling this subroutine
'' PRI writeRegister(activeRegister, address) : 
DAT writeRegister       mov     oByte, #WRITE_REGISTER
                        add     oByte, activeRegister
                        andn    outa, csnMask           ' actively communicating
                        call    #shiftOut
                        rdbyte  oByte, address    
                        call    #shiftOut          ' write byte from hub to Nordic
                        or      outa, csnMask           ' stop communicating
                        
writeRegister_ret       ret
'------------------------------------------------------------------------------------------------------------------------------
'' value of oByte should be set before calling this subroutine
DAT shiftOut            mov     bitCountdown, bits
                        ror     oByte, #8              ' this could be bits instead of #8                                                       
:bit                    shl     oByte, #1 wc         ' shift off the bit to send

                        muxc    outa, mosiMask        
                        or      outa, clockMask
                        andn    outa, clockMask
                        djnz    bitCountdown, #:bit
                     
                        
shiftOut_ret            ret     
                                          
'------------------------------------------------------------------------------------------------------------------------------
'' value of oByte should be set before calling this subroutine
'' returns value of iByte
DAT shiftIO             mov     bitCountdown, bits
                        
                        ror     oByte, #8              ' this could be bits instead of #8  ****                                                     
:bit                    shl     oByte, #1 wc         ' shift off the bit to send
                        muxc    outa, mosiMask        
                        or      outa, clockMask
                        test    misoMask, ina wc
                        rcl     iByte, #1            
                        andn    outa, clockMask
                        djnz    bitCountdown, #:bit
                     
                        
shiftIO_ret             ret                                          
'------------------------------------------------------------------------------------------------------------------------------
'' returns value of iByte 
DAT shiftIn             mov     bitCountdown, bits          
                        
:bit                    or      outa, clockMask
                        test    misoMask, ina wc        
                        rcl     iByte, #1            
                        andn    outa, clockMask
                        djnz    bitCountdown, #:Bit        
shiftIn_ret             ret 
'------------------------------------------------------------------------------------------------------------------------------
triggerDelayCode        mov     tempDelay, cnt
                        add     tempDelay, triggerDelay
                        waitcnt tempDelay, zero
                        
DAT targetIoOn          andn    outa, triggerTargetMask         ' add delay if any here                        
                        mov     targetIoFlag, #1
                        jmp     #returnPoint
'------------------------------------------------------------------------------------------------------------------------------
DAT targetIoRoutineOff  tjz     targetIoFlag, #targetIoRoutineOff_ret
                        mov     targetIoFlag, zero
                        add     targetIoCountCog, #1
                        wrlong  targetIoCountCog, targetIoCountPtr
                        mov     tempDelay, cnt
                        add     tempDelay, targetIoOffDelay
                        waitcnt tempDelay, zero
                        or      outa, triggerTargetMask
targetIoRoutineOff_ret  ret                       
'------------------------------------------------------------------------------------------------------------------------------
'' value of ledCountdown should be set before calling this subroutine
'  This subrutine is currently not used.  It may be called as a debug aid when needed.
DAT debugLed            or      dira, ledMask
                        xor     outa, ledMask
                        mov     tempDelay, cnt
                        add     tempDelay, quarterSecond
                        waitcnt tempDelay, zero
                        djnz    ledCountdown, #debugLed
debugLed_ret            ret
'------------------------------------------------------------------------------------------------------------------------------
DAT fixTriggerCheck     mov     loop, nop_      ' lazy way to write self modifying code
                        mov     loopP1, nop_
                        nop
                        'jmp     #loop
                        jmp     #configure_
'------------------------------------------------------------------------------------------------------------------------------
DAT {slowDelayRoutine    mov     tempDelay, cnt
                        add     tempDelay, slowDelay
                        waitcnt tempDelay, zero
slowDelayRoutine_ret    ret     }
{
########################### Assembly variables ###########################
}
nop_                    nop
zero                    long    0                       ''constants
negativeOne             long    -1
quarterSecond           long    20_000_000
targetIoOffDelay        long    4_000           ' 0.05ms
ceToggleDelay           long    800
slowDelay               long    45 '40 too short in 5p'45 works in 5p
targetDataPipeCog       long    (TRIGGER_DATAPIPE << 1) + %0100_0000 ' ** Does this need to change based on other settings?
                                                        ''temp variables
cogCmd                  long    0                       ''     Used for DataPin mask     and     COG shutdown 

clockMask               long    1
misoMask                long    1
mosiMask                long    1
irqMask                 long    1
triggerMask             long    1 ' trigger on master device
csnMask                 long    1
ceMask                  long    1
triggerTargetMask       long    1 ' triggerTarget on slave device
ledMask                 long    %0100_0000_0000_0000_0000

bits                    long    8
inBytePtr               long    0
outBytePtr              long    0
flagPtr                 long    0
jumpMode                long    0
commandPtr              long    0
parentCommandPtr        long    0
rxPayloadPtrLoc         long    0
txPayloadPtrLoc         long    0
bufferSizePtr           long    0
rxCountPtr              long    0  
rxCountsPtr             long    0  
triggerCountPtr         long    0  
targetIoCountPtr        long    0
dataPipePtr             long    0
dataPipeCog             long    0
statusPtr               long    0
statusCog               long    0
txAddressLocationPtr    long    0
txAddressLocationCog    long    0
dataStatus              long    0
dataConfigTx            long    0
dataPayload             long    0
dataConfigRx            long    0
configPtr               long    0
addressSize             long    0
payloadSize             long    0
statusReset             long    0
rxCountCog              long    0   
rxCountsCog             long    0   
triggerCountCog         long    0 
targetIoCountCog        long    0
targetIoFlag            long    0
pRxPipe                 long    6               ' not currently used
'triggerDelay              long    10
triggerDelay            long    20_000_000
               
tempDelay               res     1
temp                    res     1
oByte                   res     1                      
iByte                   res     1
bitCountdown            res     1                       
byteCountdown           res     1                     
address                 res     1                      
configData              res     1
activeRegister          res     1
rxByte                  res     1
'pipe0Payload            res     1
rxPipe                  res     1
rxPayload               res     1
'debugAddress            res     1
'debugAddressL           res     1 
'readAddress             res     1 
'readAddressL            res     1 
ledCountdown            res     1
'debugData               res     1
t1                      res     1
t2                      res     1
                        fit
                       